# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from collections import OrderedDict
from omniconf.exceptions import UnknownSettingError, UnconfiguredSettingError
from omniconf.setting import DEFAULT_REGISTRY as SETTING_REGISTRY
import ast


def unrepr(src, _type):
    """
    Returns an interpreted value based on ``src``. If ``source`` is already an
    instance of ``_type``, no interpretation is performed.
    """
    if isinstance(src, _type):
        return src
    if not src:
        return src
    return ast.literal_eval(src)


class ConfigRegistry(object):
    """
    A registry of Configured values for a :class:`.SettingRegistry`.
    """
    def __init__(self, setting_registry=None):
        if not setting_registry:
            self.settings = SETTING_REGISTRY
        else:
            self.settings = setting_registry
        self.clear()

    def clear(self):
        self.registry = OrderedDict()

    def set(self, key, value):
        """
        Configures the value for the given key. The value will be converted to
        the type defined in the :class:`.Setting`, by calling the type as a
        function with the value as the only argument. Trying to configure a
        value under an unknown key will result in an UnknownSettingError.
        """
        if not self.settings.has(key):
            raise UnknownSettingError("Trying to configure unregistered key "
                                      "{0}".format(key))
        setting = self.settings.get(key)

        if setting.type in (list, dict, tuple, bool):
            self.registry[key] = unrepr(value, setting.type)
        else:
            self.registry[key] = setting.type(value)

    def has(self, key):
        """
        Checks if a value has been configured for the given key, or if a
        default value is present.
        """
        if key in self.registry:
            return True
        elif self.settings.has(key) and \
                self.settings.get(key).default is not None:
            return True
        return False

    def get(self, key):
        """
        Returns the configured value for the given key, or the default value if
        the key was not configured.
        """
        if key in self.registry:
            return self.registry[key]
        elif self.settings.has(key) and \
                self.settings.get(key).default is not None:
            return self.settings.get(key).default
        elif self.settings.has(key) and not self.settings.get(key).required:
            return None
        raise UnconfiguredSettingError("No value or default available for {0}"
                                       .format(key))

    def list(self):
        """
        Returns all configured values as a dict.
        """
        return self.registry

    def unset(self, key):
        """
        Removes the value for a given key from the registry.
        """
        if key in self.registry:
            del self.registry[key]

    def load(self, backends):
        """
        Attempt to configure all settings defined in the
        :class:`.SettingRegistry` using the provided backends. If a setting
        was attempting to load, and no value found and no default was set, an
        UnconfiguredSettingError is raised.
        """
        for backend in backends:
            for setting, value in backend.get_values(self.settings.list()):
                if setting.key in self.registry:
                    continue
                try:
                    self.set(setting.key, value)
                except ValueError as ve:
                    raise ValueError("An invalid value was specified for "
                                     "{0}: {1}".format(setting.key, ve))

        missing_settings = []
        for setting in self.settings.list():
            if not self.has(setting.key) and setting.required:
                missing_settings.append(setting.key)
        if missing_settings:
            raise UnconfiguredSettingError("No value was configured for: {0}"
                                           .format(", "
                                                   .join(missing_settings)))


DEFAULT_REGISTRY = ConfigRegistry()
"""
Global :class:`.ConfigRegistry` which will be used when no specific
:class:`.ConfigRegistry` is defined.
"""


def config(key, registry=None):
    """
    Retrieves the configured value for a given key. If no specific registry is
    specified, the value will be retrieved from the default
    :class:`.ConfigRegistry`.
    """
    if not registry:
        registry = DEFAULT_REGISTRY

    return registry.get(key)



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.


class ConfigBackend(object):
    """
    Defines a configuration backend, which provides configuration values
    based on keys.
    """

    def __init__(self, conf=None):
        self.config = conf

    @classmethod
    def autodetect_settings(cls, autoconfigure_prefix):
        """
        Returns a tuple of :class:`.Setting` objects, that are required for
        :func:`autoconfigure` to complete successfully.
        """
        return ()

    @classmethod
    def autoconfigure(cls, conf, autoconfigure_prefix):
        """
        Called with a :class:`.ConfigRegistry`, the result of this method must
        be either a new instance of this class, or :any:`None`. This method
        is automatically called during the autoconfigure phase.
        """
        raise NotImplementedError("This method must be implemented")

    def get_value(self, setting):
        """
        Retrieves the value for the given :class:`.Setting`.
        """
        section = self.config
        for _key in setting.key.split("."):
            section = section[_key]
        return section

    def get_values(self, settings):
        """
        Retrieves a list of :class:`.Setting`s all at once. Values are returned
        as a list of tuples containing the :class:`.Setting` and value.
        """
        values = []
        for setting in settings:
            try:
                values.append((setting, self.get_value(setting)))
            except KeyError:
                pass
        return values



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from mock import Mock
import unittest
from omniconf.setting import SettingRegistry, Setting, setting, \
                                DEFAULT_REGISTRY


class TestSettingsRegistry(unittest.TestCase):
    def setUp(self):
        self.registry = SettingRegistry()
        self.setting = Setting(key="foo", _type=str)
        self.registry.add(self.setting)

    def test_setting_registry_key_with_string(self):
        self.assertEqual('foo', self.registry._key('foo'))
        self.assertEqual('section.foo', self.registry._key('section.foo'))
        self.assertEqual('section.subsection.foo',
                         self.registry._key('section.subsection.foo'))

    def test_setting_registry_key_with_object(self):
        obj = Mock()
        obj.key = "foo"
        self.assertEqual('foo', self.registry._key(obj))
        obj.key = "section.foo"
        self.assertEqual('section.foo', self.registry._key(obj))
        obj.key = "section.subsection.foo"
        self.assertEqual('section.subsection.foo', self.registry._key(obj))

    def test_setting_registry_clear(self):
        self.assertEqual(len(self.registry.registry), 1)
        self.registry.clear()
        self.assertEqual(len(self.registry.registry), 0)

    def test_setting_registry_add(self):
        self.assertIn(self.setting, self.registry.registry.values())

    def test_setting_registry_has(self):
        self.assertTrue(self.registry.has("foo"))

    def test_setting_registry_get(self):
        self.assertEqual(self.registry.get("foo"), self.setting)

    def test_setting_registry_keys(self):
        self.assertEqual(self.registry.keys(),
                         list(self.registry.registry.keys()))

    def test_setting_registry_list(self):
        self.assertEqual(self.registry.list(),
                         list(self.registry.registry.values()))

    def test_setting_registry_remove(self):
        self.registry.remove(self.setting)
        self.assertNotIn(self.setting, self.registry.registry.values())


class TestSettingMethod(unittest.TestCase):
    def setUp(self):
        self.registry = SettingRegistry()

    def test_setting_method(self):
        _setting = setting("foo", registry=self.registry)
        self.assertTrue(len(self.registry.registry.values()) == 1)
        self.assertIn(_setting, self.registry.registry.values())

    def test_setting_method_with_default_registry(self):
        _setting = setting("foo")
        self.assertIn(_setting, DEFAULT_REGISTRY.registry.values())
        self.assertNotEqual(id(self.registry), id(DEFAULT_REGISTRY))

        DEFAULT_REGISTRY.remove(_setting)



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from omniconf.backends.generic import ConfigBackend
from omniconf.loader import autoconfigure_backends, omniconf_load
from omniconf.setting import Setting
from mock import Mock, patch
import unittest

autodetection_mock = Mock(autospec=ConfigBackend)
autodetection_mock().get_values.return_value = []

autoconfigure_mock = Mock(autospec=ConfigBackend)
autoconfigure_mock().get_values.return_value = []
autoconfigure_mock.autodetect_settings.return_value = [
    Setting("omniconf.foo", _type=str)]
autoconfigure_mock.autoconfigure.return_value = autoconfigure_mock


class TestLoader(unittest.TestCase):
    def setUp(self):
        autodetection_mock.reset_mock()
        autoconfigure_mock.reset_mock()

    @patch("omniconf.loader.autodetection_backends", new=[autodetection_mock])
    @patch("omniconf.loader.available_backends", new=[autoconfigure_mock])
    def test_autoconfigure_backends(self):
        prefix = "testconf"
        configured_backends = autoconfigure_backends(prefix)

        autoconfigure_mock.autodetect_settings.assert_called_once_with(prefix)
        autodetection_mock.assert_called_once_with()
        self.assertEqual(configured_backends, [autoconfigure_mock])

    @patch("omniconf.loader.autodetection_backends", new=[autodetection_mock])
    @patch("omniconf.loader.available_backends", new=[autoconfigure_mock])
    @patch("omniconf.loader.CONFIG_REGISTRY")
    def test_omniconf_load(self, registry_mock):
        omniconf_load()
        registry_mock.load.assert_called_once_with([autoconfigure_mock])

    @patch("omniconf.loader.CONFIG_REGISTRY")
    def test_omniconf_load_with_backends(self, registry_mock):
        omniconf_load(backends="FOO")
        registry_mock.load.assert_called_once_with("FOO")



# Copyright (c) 2019 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

try:
    from collections import abc
except ImportError:  # pragma: nocover
    import collections as abc

try:
    string_types = (str, unicode)
except NameError:  # pragma: nocover
    string_types = (str, bytes)


class SeparatorSequence(abc.Sequence):
    """
    Splits the given string using the given separator, and provides a
    the result with a read-only Sequence interface.
    """
    def __init__(self, string, separator):
        self.lst = string.split(separator)

    def __getitem__(self, key):
        return self.lst.__getitem__(key)

    def __len__(self):
        return self.lst.__len__()

    def __str__(self):
        return self.lst.__str__()

    __unicode__ = __str__

    def __repr__(self):
        return self.lst.__repr__()


def separator_sequence(separator):
    """
    Returns a function that parses a string value, separates it into parts and
    stores it as a read-only sequence:

    .. code-block:: python

        parser = separator_sequence(",")
        print parser("a,b,c")
        # ['a', 'b', 'c']

    If the input value is already a sequence (but not a string), the value is
    returned as is. The sequence is an instance of :class:`.SeparatorSequence`,
    and can be used as one would normally use a (read-only) tuple or list.
    """
    def factory(value):
        if isinstance(value, abc.Sequence) and \
                not isinstance(value, string_types):
            return value
        return SeparatorSequence(value, separator)
    return factory


def string_or_false(value):
    """
    Returns the given value as-is, unless the values equals "False". In that
    case, boolean False is returned.
    """
    if value == "False":
        return False
    return value


def string_bool(value):
    """
    Returns False if the value is Falsish or "False", True if value is "True",
    or the original value otherwise.
    """
    if isinstance(value, bool):
        return value
    if not value or value == "False":
        return False
    if value == "True":
        return True
    return value


def enum(values):
    """
    Returns the original value if it is present in values, otherwise raises a
    RuntimeError.

    .. code-block:: python

        enum_func = enum(["foo", "bar"])
        print enum_func("foo")
        # "foo"
        print enum_func("baz")
        # ...
        # RuntimeError: Invalid value specified, must be one of: foo, bar

    """
    def factory(value):
        if value not in values:
            raise RuntimeError("Invalid value specified, must be one of: {}"
                               .format(", ".join(values)))
        return value
    return factory



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from omniconf.backends import available_backends
from omniconf.backends.env import EnvBackend
from omniconf.setting import Setting
from mock import patch
import nose.tools
import os

ENV_FILE = {
    "COLORTERM": "gnome-terminal",
    "DESKTOP_MODE": "1",
    "DESKTOP_SESSION": "ubuntu",
    "DISPLAY": ":0",
    "EDITOR": "vim",
    "HOME": "/home/user/workspace/omniconf",
    "LANG": "en_US.UTF-8",
    "LANGUAGE": "en_US:",
    "LC_ADDRESS": "nl_NL.UTF-8",
    "LC_IDENTIFICATION": "nl_NL.UTF-8",
    "LC_MEASUREMENT": "nl_NL.UTF-8",
    "LC_MONETARY": "nl_NL.UTF-8",
    "LC_NAME": "nl_NL.UTF-8",
    "LC_NUMERIC": "nl_NL.UTF-8",
    "LC_PAPER": "nl_NL.UTF-8",
    "LC_TELEPHONE": "nl_NL.UTF-8",
    "LC_TIME": "nl_NL.UTF-8",
    "OLDPWD": "/home/user",
    "PATH": "/foo/bar/baz",
    "PWD": "/home/user",
    "SHELL": "/bin/bash",
    "TERM": "xterm",
    "USER": "user",
    "_": "/home/user/workspace/omniconf/vendor/bin/python",


    "TEST_FOO": "bar",
    "TEST_SECTION_BAR": "baz",
    "TEST_SECTION_SUBSECTION_BAZ": "foo"
}

CONFIGS = [
    ("foo", "bar", None),
    ("section.bar", "baz", None),
    ("section.subsection.baz", "foo", None),
    ("", None, KeyError),
    ("section", None, KeyError),
    ("unknown", None, KeyError)
]


def test_env_backend_in_available_backends():
    nose.tools.assert_in(EnvBackend, available_backends)


def test_env_backend_autoconfigure():
    prefix = "testconf"
    backend = EnvBackend.autoconfigure(
        {"{0}.prefix".format(prefix): "bla"}, prefix)
    nose.tools.assert_is_instance(backend, EnvBackend)
    nose.tools.assert_equal(backend.prefix, "bla")


def test_env_backend_get_value():
    for key, value, sideeffect in CONFIGS:
        yield _test_get_value, key, value, sideeffect


def _test_get_value(key, value, sideeffect):
    with patch.dict(os.environ, ENV_FILE):
        backend = EnvBackend(prefix="TEST")
        setting = Setting(key=key, _type=str)
        if sideeffect:
            with nose.tools.assert_raises(sideeffect):
                backend.get_value(setting)
        else:
            nose.tools.assert_equal(backend.get_value(setting), value)



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

import unittest
from omniconf.backends.generic import ConfigBackend
from omniconf.config import ConfigRegistry, config, DEFAULT_REGISTRY, \
                            SETTING_REGISTRY
from omniconf.exceptions import UnknownSettingError, \
                                UnconfiguredSettingError
from omniconf.setting import SettingRegistry, Setting
from mock import Mock
import nose.tools


class TestConfigRegistry(unittest.TestCase):
    def setUp(self):
        self.test_settings = [
            Setting("key", _type=str, required=True),
            Setting("default", _type=str, default="present")
        ]
        self.setting_registry = SettingRegistry()
        for setting in self.test_settings:
            self.setting_registry.add(setting)
        self.config_registry = ConfigRegistry(
                                setting_registry=self.setting_registry)

    def test_config_registry_clear(self):
        self.assertEqual(len(self.setting_registry.registry), 2)
        self.setting_registry.clear()
        self.assertEqual(len(self.setting_registry.registry), 0)

    def test_config_registry_set_without_setting(self):
        with self.assertRaises(UnknownSettingError):
            self.config_registry.set("nope", "value")

    def test_config_registry_set_with_setting(self):
        self.config_registry.set("key", "value")
        self.assertEqual(self.config_registry.registry['key'], "value")

    def test_config_registry_has(self):
        self.assertFalse(self.config_registry.has("key"))
        self.config_registry.set("key", "added")
        self.assertTrue(self.config_registry.has("key"))

    def test_config_registry_has_with_default(self):
        self.assertTrue(self.config_registry.has("default"))

    def test_config_registry_get_without_config(self):
        with self.assertRaises(UnconfiguredSettingError):
            self.config_registry.get("nope")

    def test_config_registry_get(self):
        self.config_registry.set("key", "value")
        self.assertEqual(self.config_registry.get("key"), "value")

    def test_config_registry_get_with_default(self):
        self.assertEqual(self.config_registry.get("default"), "present")

    def test_config_registry_get_not_required_no_value(self):
        self.setting_registry.add(Setting("unneeded", _type=str))
        self.assertIs(self.config_registry.get("unneeded"), None)

    def test_config_registry_list(self):
        self.assertEqual(self.config_registry.list(),
                         self.config_registry.registry)

    def test_config_registry_unset(self):
        self.config_registry.set("key", "soon")
        self.assertEqual(self.config_registry.get("key"), "soon")
        self.config_registry.unset("key")
        with self.assertRaises(UnconfiguredSettingError):
            self.config_registry.get("key")

    def test_config_registry_load(self):
        mock_backend = Mock(autospec=ConfigBackend)
        mock_backend.get_values.return_value = [(s, "value")
                                                for s in self.test_settings]
        self.config_registry.load([mock_backend])

        mock_backend.get_values.assert_called_once_with(self.test_settings)
        self.assertEqual(self.config_registry.get("key"), "value")
        self.assertEqual(self.config_registry.get("default"), "value")

    def test_config_registry_load_with_previous_values(self):
        mock_backend = Mock(autospec=ConfigBackend)
        mock_backend.get_values.return_value = [(s, "value")
                                                for s in self.test_settings]
        self.config_registry.set("key", "other")
        self.config_registry.set("default", "other")

        self.config_registry.load([mock_backend])

        self.assertEqual(self.config_registry.get("key"), "other")
        self.assertEqual(self.config_registry.get("default"), "other")

    def test_config_registry_load_with_unavailable_values(self):
        mock_backend = Mock(autospec=ConfigBackend)
        mock_backend.get_values.return_value = []

        with self.assertRaises(UnconfiguredSettingError):
            self.config_registry.load([mock_backend])

    def test_config_registry_load_value_error(self):
        int_setting = Setting(key="foo", _type=int)
        self.setting_registry.add(int_setting)

        mock_backend = Mock(autospec=ConfigBackend)
        mock_backend.get_values.return_value = [
            (int_setting, "bar")]

        with self.assertRaises(ValueError):
            self.config_registry.load([mock_backend])


class TestConfigMethod(unittest.TestCase):
    def setUp(self):
        self.setting_registry = SettingRegistry()
        self.setting_registry.add(Setting("key", _type=str, required=True))
        self.setting_registry.add(Setting("default", _type=str,
                                          default="present"))
        self.config_registry = ConfigRegistry(
                                setting_registry=self.setting_registry)

    def test_config_method_without_config(self):
        with self.assertRaises(UnconfiguredSettingError):
            config("key", registry=self.config_registry)

    def test_config_method(self):
        self.config_registry.set("key", "value")
        self.assertEqual(config("key", registry=self.config_registry), "value")

    def test_config_method_with_default(self):
        self.assertEqual(config("default", registry=self.config_registry),
                         "present")

    def test_config_method_with_default_registry(self):
        with self.assertRaises(UnconfiguredSettingError):
            config("foo")

        _setting = Setting(key="foo", _type=str)
        SETTING_REGISTRY.add(_setting)
        DEFAULT_REGISTRY.set("foo", "bar")
        self.assertEqual(config("foo"), "bar")

        DEFAULT_REGISTRY.unset("foo")
        SETTING_REGISTRY.remove(_setting)


VALUE_TESTS = [
    # Normal values
    ("foobar", "foobar", str, None),
    (123456, "123456", str, None),
    ("1234", 1234, int, None),
    ("123.456", 123.456, float, None),
    ("['a', 'b', '1', 'c']", ['a', 'b', '1', 'c'], list, None),
    ("('a', 'b', '1', 'c')", ('a', 'b', '1', 'c'), tuple, None),
    ("{'a': 'b', 'foo': 'bar'}", {'a': 'b', 'foo': 'bar'}, dict, None),
    ("False", False, bool, None),
    ("True", True, bool, None),
    (False, False, bool, None),

    # Value is False codepath
    (False, False, list, None),

    # Exception codepath
    ("foobar", None, int, ValueError),
    ("foobar", None, float, ValueError),
    ("foobar", None, list, ValueError),
    ("foobar", None, dict, ValueError)
]


def test_config_registry_set_value_conversion():
    for _in, _out, _type, sideeffect in VALUE_TESTS:
        yield _test_set_value, _in, _out, _type, sideeffect


def _test_set_value(_in, _out, _type, sideeffect):
    settings = SettingRegistry()
    settings.add(Setting("test.value", _type=_type))

    configs = ConfigRegistry(setting_registry=settings)

    if sideeffect:
        with nose.tools.assert_raises(sideeffect):
            configs.set("test.value", _in)
    else:
        configs.set("test.value", _in)
        nose.tools.assert_equal(configs.get("test.value"), _out)



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from omniconf.types import separator_sequence, string_or_false, string_bool, \
    enum
import nose.tools


SEPARATOR_SEQUENCES = [
    ("", ",", [""]),
    ("a", ",", ["a"]),
    ("a,b,c", ",", ["a", "b", "c"]),
    ("foo,bar;baz;", ";", ["foo,bar", "baz", ""]),
    (["foo", "bar"], ",", ["foo", "bar"]),
    ([], ",", [])
]


def _test_separator_sequence(_in, _sep, _out):
    seq = separator_sequence(_sep)(_in)
    nose.tools.assert_sequence_equal(seq, _out)
    nose.tools.assert_equal(seq.__str__(), _out.__str__())


def test_separator_sequence():
    for _in, _sep, _out in SEPARATOR_SEQUENCES:
        yield _test_separator_sequence, _in, _sep, _out


STRING_OR_FALSE = [
    ("foo", "foo"),
    ("bar", "bar"),
    (123, 123),
    ("False", False),
    (None, None)
]


def _test_string_or_false(_in, _out):
    nose.tools.assert_equal(string_or_false(_in), _out)


def test_string_or_false():
    for _in, _out in STRING_OR_FALSE:
        yield _test_string_or_false, _in, _out


STRING_BOOL = [
    (0, False),
    ("", False),
    ([], False),
    ({}, False),
    ("False", False),
    (False, False),
    ("True", True),
    (True, True),
    ("0", "0"),
    (1, 1),
    ("1", "1")
]


def _test_string_bool(_in, _out):
    nose.tools.assert_equal(string_bool(_in), _out)


def test_string_bool():
    for _in, _out in STRING_BOOL:
        yield _test_string_bool, _in, _out


ENUM = enum(["foo", "bar", "baz"])
ENUMS = [
    ("foo", "foo", None),
    ("bar", "bar", None),
    ("baz", "baz", None),
    ("fun", None, RuntimeError)
]


def _test_enum(_in, _out, _exc):
    if _exc:
        with nose.tools.assert_raises(_exc):
            ENUM(_in)
    else:
        nose.tools.assert_equal(ENUM(_in), _out)


def test_enum():
    for _in, _out, _exc in ENUMS:
        yield _test_enum, _in, _out, _exc



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from collections import OrderedDict


class SettingRegistry(object):
    """
    A registry of defined :class:`Setting` objects.
    """
    def __init__(self):
        self.registry = OrderedDict()

    def _key(self, setting):
        if isinstance(setting, str):
            return setting
        else:
            return setting.key

    def clear(self):
        self.registry = OrderedDict()

    def add(self, setting):
        """
        Register a Setting.
        """
        self.registry[setting.key] = setting
        return setting

    def has(self, key):
        """
        Check if a Setting has been registered under the specified key.
        """
        return key in self.registry

    def get(self, key):
        """
        Retrieves the Setting for the given key.
        """
        return self.registry[key]

    def keys(self):
        """
        Returns the registered keys.
        """
        return list(self.registry.keys())

    def list(self):
        """
        Returns the configured Settings as a list.
        """
        return list(self.registry.values())

    def remove(self, setting):
        """
        Removes the Setting with the given key.
        """
        del self.registry[self._key(setting)]


DEFAULT_REGISTRY = SettingRegistry()
"""
Global :class:`.SettingRegistry` which will be used when no specific
:class:`.SettingRegistry` is defined.
"""


class Setting(object):
    """
    A :class:`.Setting` is registered under a specific key and with a specific
    type (:class:`str`, :class:`dict`, :class:`list`, etc). A default may also
    be specified, which allows a config to be returned without a value being
    specifically defined (also see :class:`.ConfigRegistry`). A
    :class:`.Setting` may be marked as required, which will cause an exception
    to be thrown when no value is found when autoconfiguring. A help message
    may be specified for documentation purposes.
    """
    def __init__(self, key, _type, required=False, default=None, help=None):
        self.key = key
        self.type = _type
        self.required = required
        self.default = default
        self.help = help


def setting(key, _type=str, required=False, default=None, help=None,
            registry=None):
    """
    Register a new :class:`.Setting` with the given key.
    """
    if not registry:
        registry = DEFAULT_REGISTRY

    return registry.add(Setting(key, _type, required, default, help))



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from omniconf.backends.generic import ConfigBackend
from omniconf.setting import Setting
import nose.tools


def test_config_backend_autoconfigure():
    with nose.tools.assert_raises(NotImplementedError):
        ConfigBackend.autoconfigure(None, None)


def test_config_backend_autodetect_settings():
    nose.tools.assert_equal(ConfigBackend.autodetect_settings(None), ())


def test_config_backend_get_values_no_settings():
    nose.tools.assert_equal(ConfigBackend().get_values([]), [])


def test_config_backend_get_values_missing_value():
    backend = ConfigBackend(conf={})
    setting = Setting("foo", _type=str)
    values = backend.get_values([setting])
    nose.tools.assert_equal(values, [])


def test_config_backend_get_values():
    backend = ConfigBackend(conf={"foo": "bar"})
    setting = Setting("foo", _type=str)
    values = backend.get_values([setting])
    nose.tools.assert_equal(values, [(setting, "bar")])



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from omniconf.backends import available_backends
from omniconf.backends.argparse import ArgparseBackend
from omniconf.config import ConfigRegistry
from omniconf.setting import Setting, SettingRegistry
from mock import patch
import nose.tools

ARGS_FILE = [
    "--foo", "bar",
    "--section-bar", "baz",
    "--section-subsection-baz", "foo",
    "--bool-normal", "1",
    "--bool-true",
    "--bool-false",
    "--missing-value"  # Has to be the last because we're omitting the value
]

PREFIX_ARGS_FILE = [
    "--prefix-foo", "bar",
    "--prefix-section-bar", "baz",
    "--prefix-section-subsection-baz", "foo",
    "--prefix-bool-normal", "1",
    "--prefix-bool-true",
    "--prefix-bool-false",
    "--prefix-missing-value"  # Has to be the last because we're
                              # omitting the value
]

CONFIGS = [
    (Setting(key="foo", _type=str), "bar", None),
    (Setting(key="section.bar", _type=str), "baz", None),
    (Setting(key="section.subsection.baz", _type=str), "foo", None),

    (Setting(key="", _type=str), None, KeyError),
    (Setting(key="missing.value", _type=str), None, KeyError),
    (Setting(key="missing.arg", _type=str), None, IndexError),  # Raise in test

    (Setting(key="bool.normal", _type=bool), "1", None),
    (Setting(key="bool.true", _type=bool, default=False), True, None),
    (Setting(key="bool.false", _type=bool, default=True), False, None),
    (Setting(key="bool.default.true", _type=bool, default=True), True, None),
    (Setting(key="bool.default.false", _type=bool, default=False), False, None)
]


def test_argparse_backend_in_available_backends():
    nose.tools.assert_in(ArgparseBackend, available_backends)


def test_argparse_backend_autoconfigure():
    prefix = "testconf"
    backend = ArgparseBackend.autoconfigure(
        {"{0}.prefix".format(prefix): "bar"}, prefix)
    nose.tools.assert_is_instance(backend, ArgparseBackend)
    nose.tools.assert_equal(backend.prefix, "bar")


def test_argparse_backend_get_value():
    with nose.tools.assert_raises(NotImplementedError):
        backend = ArgparseBackend()
        backend.get_value(None)


def test_argparse_backend_get_values():
    for setting, value, sideeffect in CONFIGS:
        yield _test_get_values, setting, value, sideeffect, None
        yield _test_get_values, setting, value, sideeffect, 'prefix'


def _test_get_values(setting, value, sideeffect, prefix):
    with patch('omniconf.backends.argparse.ARGPARSE_SOURCE',
               ARGS_FILE if not prefix else PREFIX_ARGS_FILE):
        backend = ArgparseBackend(prefix=prefix)
        if sideeffect:
            with nose.tools.assert_raises(sideeffect):
                setting, config = backend.get_values([setting])[0]
        else:
            setting, config = backend.get_values([setting])[0]
            nose.tools.assert_equal(config, value)


def test_mixed_flags_and_settings():
    MIXED_ARGS = [
        "--verbose", "--loud",
        "--foo-bar", "baz",
        "--bar", "buzz"
    ]

    settings = SettingRegistry()
    settings.add(Setting(key="verbose", _type=bool, default=False))
    settings.add(Setting(key="loud", _type=bool, default=True))
    settings.add(Setting(key="foo.bar", _type=str))
    settings.add(Setting(key="bar", _type=str))
    configs = ConfigRegistry(setting_registry=settings)

    with patch('omniconf.backends.argparse.ARGPARSE_SOURCE', MIXED_ARGS):
        backend = ArgparseBackend(prefix=None)
        configs.load([backend])

        nose.tools.assert_true(configs.get("verbose"))
        nose.tools.assert_false(configs.get("loud"))
        nose.tools.assert_equal(configs.get("foo.bar"), "baz")
        nose.tools.assert_equal(configs.get("bar"), "buzz")



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from collections import OrderedDict
from omniconf.backends.generic import ConfigBackend
from omniconf.keys import join_key
import argparse
import sys

ARGPARSE_SOURCE = sys.argv[1:]


def suppress_output(*args, **kwargs):  # pragma: nocover
    pass


def format_argparse_key(key, prefix=None):
    _key = key.replace(".", "-").lower()
    _prop = key.replace(".", "_").lower()
    if prefix:
        _key = "{0}-{1}".format(prefix, _key)
        _prop = "{0}_{1}".format(prefix, _prop)
    _arg = "--{0}".format(_key)

    return _key, _prop, _arg


class ArgparseBackend(ConfigBackend):
    """
    Uses the current process arguments, and allows values in it to
    be retrieved using dotted keys with a specific prefix. By default no
    prefix is assumed.
    """

    def __init__(self, conf=None, prefix=None):
        super(ArgparseBackend, self).__init__()
        self.prefix = prefix if prefix else ""

    @classmethod
    def autoconfigure(cls, conf, autoconfigure_prefix):
        return ArgparseBackend(prefix=conf.get(join_key(autoconfigure_prefix,
                                                        "prefix")))

    @classmethod
    def add_argument(cls, parser, argument, setting):
        if setting.type is bool:
            if setting.default is None:
                return parser.add_argument(argument)
            elif setting.default is True:
                return parser.add_argument(argument, action="store_false")
            else:
                return parser.add_argument(argument, action="store_true")
        return parser.add_argument(argument)

    def get_value(self, setting):
        raise NotImplementedError("get_value is not implemented for "
                                  "ArgparseBackend, use get_values instead.")

    def get_values(self, settings):
        """
        Process the given list :class:`.Setting` objects, and retrieve the
        values. Keys are converted as follows:

        * Dots are replaced by dashes (-).
        * The key is lowercased.
        * A prefix is attached to the key, if specified

        This means that a key like section.value will be queried like
        ``--prefix-section-value``. When no prefix is specified,
        ``--section-value`` is queried instead.

        Special handling is added for boolean Settings with a default
        specified, which works as follows:

        * Settings with `_type=bool` and no default will be processed
          as normal.
        * Settings with `_type=bool`, and where the default value is True will
          be specified as an argparse argument with `action=store_false`.
        * Settings with `_type=bool`, and where the default value is False will
          be specified as an argparse argument with `action=store_true`.
        """
        parser = argparse.ArgumentParser(add_help=False)

        # Disable forced output from argparse we don't want to display
        parser.print_usage = suppress_output
        parser._print_message = suppress_output

        for setting in settings:
            if not setting.key:
                raise KeyError("Empty keys are not allowed")
            _key, _prop, _arg = format_argparse_key(setting.key, self.prefix)
            ArgparseBackend.add_argument(parser, _arg, setting)

        try:
            args = parser.parse_known_args(args=ARGPARSE_SOURCE)[0]
        except SystemExit:
            raise KeyError("Error parsing value for {0}".format(setting.key))

        arguments = vars(args)
        values = []
        for setting in settings:
            _, _prop, _ = format_argparse_key(setting.key, self.prefix)
            if _prop in arguments and arguments[_prop] is not None:
                values.append((setting, arguments[_prop]))

        return values


class ArgparseUsageInformation(object):
    """
    Formats the settings in :class:`.SettingRegistry`, and formats a typical
    Unix usage message for output to console.

    If no name is specified, the value in `sys.argv[0]` will be used.
    Additionally, a custom header and footer can be specified using the
    `top_message` and `bottom_message` parameters.
    """
    def __init__(self, setting_registry, name=None, top_message=None,
                 bottom_message=None):
        self.registry = setting_registry
        self.name = name if name else sys.argv[0]
        self.top_message = top_message
        self.bottom_message = bottom_message

    def _sortable_key(self, key):
        """
        Prefixes keys without a section with an underscore. This will cause
        options like `verbose` to be mentioned before all other options.
        """
        if "." not in key:
            key = "_." + key
        return key

    def _short_metavar_name(self, prop):
        """
        Creates a shortened metavar based on the name of a property. The passed
        property is first uppercased, and split on underscores. Then the first
        letter of each part is used. An property named "foo_bar_baz" will
        result in a metavar named "FBB".
        """
        prop = prop.upper()
        if "_" in prop:
            parts = prop.split("_")
            prop = "".join([part[0] for part in parts])
        elif len(prop) > 3:
            prop = prop[0:3]
        return prop

    def group_settings(self):
        """
        Orders all registered :class:`.Setting` objects by key, and then groups
        then based on the first part of the key.
        """
        keys = sorted(self.registry.keys(),
                      key=lambda x: self._sortable_key(x))
        groups = OrderedDict()
        for key in keys:
            parts = key.split(".")
            group = parts[0] if len(parts) > 1 else "_"
            if group not in groups:
                groups[group] = []
            groups[group].append(key)
        return keys, groups

    def print_usage(self, out=None):
        """
        Leverages :mod:`argparse` to create human readable usage information.
        """
        keys, groups = self.group_settings()
        usage_argparse = argparse.ArgumentParser(
            description=self.top_message, prog=self.name, add_help=False)

        for group, keys in groups.items():
            if group == "_":
                group_argparse = usage_argparse
            else:
                group_argparse = usage_argparse.add_argument_group(group)

            for key in keys:
                _, _prop, _arg = format_argparse_key(key)
                setting = self.registry.get(key)
                argument = ArgparseBackend.add_argument(
                    parser=group_argparse,
                    argument=_arg,
                    setting=setting
                )
                argument.default = setting.default
                argument.type = setting.type
                argument.metavar = self._short_metavar_name(_prop)
                argument.required = setting.required
                argument.help = setting.help

        usage_argparse.print_help(file=out)

    @classmethod
    def check_flag(cls, flags):
        """
        Convenience method to quickly check if the specified flags were
        provided on the command line.
        """
        flag_argparse = argparse.ArgumentParser(add_help=False)
        flag_argparse.add_argument(*flags, dest="flag",
                                   action="store_true")

        return flag_argparse.parse_known_args(args=ARGPARSE_SOURCE)[0].flag



# Copyright (c) 2020 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
from omniconf.backends import available_backends
from omniconf.backends.toml import TomlBackend
from omniconf.config import ConfigRegistry
from omniconf.setting import SettingRegistry, Setting
from mock import mock_open
try:
    from io import StringIO
except ImportError:  # pragma: nocover
    from StringIO import StringIO
import nose.tools

TOML_FILE = """
foo = "bar"
dotted.keys.are = "supported"

[section]
bar = "baz"

[section.subsection]
baz = "foo"

[bar]
sub = "bar-sub-value"
"""

CONFIGS = [
    ("foo", "bar", None),
    ("dotted.keys.are", "supported", None),
    ("section.bar", "baz", None),
    ("section.subsection.baz", "foo", None),
    ("", None, KeyError),
    ("section", {"bar": "baz", "subsection": {"baz": "foo"}}, None),
    ("unknown", None, KeyError),
    ("bar.sub", "bar-sub-value", None)
]


def test_toml_backend_in_available_backends():
    nose.tools.assert_in(TomlBackend, available_backends)


def test_toml_backend_autoconfigure():
    from omniconf.backends import toml as omniconf_backend_toml
    omniconf_backend_toml.open = mock_open(read_data=TOML_FILE)
    prefix = "testconf"
    settings = SettingRegistry()
    settings.add(TomlBackend.autodetect_settings(prefix)[0])
    conf = ConfigRegistry(setting_registry=settings)

    backend = TomlBackend.autoconfigure(conf, prefix)
    nose.tools.assert_is(backend, None)

    conf.set("{0}.toml.filename".format(prefix), "bar")
    backend = TomlBackend.autoconfigure(conf, prefix)
    nose.tools.assert_is_instance(backend, TomlBackend)


def test_toml_backend_get_value():
    for key, value, sideeffect in CONFIGS:
        yield _test_get_value, key, value, sideeffect


def _test_get_value(key, value, sideeffect):
    f = StringIO(TOML_FILE)
    backend = TomlBackend(f)
    setting = Setting(key=key, _type=str)
    if sideeffect:
        with nose.tools.assert_raises(sideeffect):
            backend.get_value(setting)
    else:
        nose.tools.assert_equal(backend.get_value(setting), value)



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from omniconf.backends.argparse import ArgparseBackend
from omniconf.backends.env import EnvBackend
from omniconf.backends.json import JsonBackend

available_backends = []

try:
    from omniconf.backends.configobj import ConfigObjBackend
    available_backends.append(ConfigObjBackend)
except ImportError:  # pragma: nocover
    pass

try:
    from omniconf.backends.yaml import YamlBackend
    available_backends.append(YamlBackend)
except ImportError:  # pragma: nocover
    pass

try:
    from omniconf.backends.toml import TomlBackend
    available_backends.append(TomlBackend)
except ImportError:  # pragma: nocover
    pass

try:
    from omniconf.backends.vault import VaultBackend
    available_backends.append(VaultBackend)
except ImportError:  # pragma: nocover
    pass

available_backends += [JsonBackend, EnvBackend, ArgparseBackend]
autodetection_backends = [EnvBackend, ArgparseBackend]


__all__ = [available_backends, autodetection_backends]



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from collections import OrderedDict
from mock import patch
from omniconf.backends.argparse import ArgparseUsageInformation, \
                                        format_argparse_key
from omniconf.setting import SettingRegistry, Setting
import nose.tools
import unittest

try:
    from StringIO import StringIO
except ImportError:  # pragma: nocover
    from io import StringIO


SETTINGS = [
    Setting("module1.str", _type=str, help="Optional str argument."),
    Setting("module1.int", _type=int, help="Optional int argument."),
    Setting("module1.required", _type=str, required=True,
            help="Required str argument."),
    Setting("module1.default", _type=str, default="foo",
            help="Optional str argument with default."),
    Setting("module2.setting", _type=str, help="Optional str argument."),
    Setting("module2.bool_flag", _type=str, default=False,
            help="Optional boolean flag with default."),
    Setting("zmodule", _type=str, help="Optional, top-level str argument."),
    Setting("xmodule.setting", _type=str, help="Optional str argument.")
]

ORDERED_SETTINGS = [
    "zmodule",
    "module1.default",
    "module1.int",
    "module1.required",
    "module1.str",
    "module2.bool_flag",
    "module2.setting",
    "xmodule.setting"
]

GROUPED_SETTINGS = OrderedDict((
    ("_", ["zmodule"]),
    ("module1", ["module1.default", "module1.int", "module1.required",
                 "module1.str"]),
    ("module2", ["module2.bool_flag", "module2.setting"]),
    ("xmodule", ["xmodule.setting"])
))


class TestArgparseUsageInformation(unittest.TestCase):
    def setUp(self):
        registry = SettingRegistry()
        for setting in SETTINGS:
            registry.add(setting)
        self.registry = registry
        self.usage = ArgparseUsageInformation(registry)

    def test_argparse_usage_information_group_settings(self):
        keys, groups = self.usage.group_settings()
        nose.tools.assert_equal(keys, ORDERED_SETTINGS)
        nose.tools.assert_equal(groups, GROUPED_SETTINGS)

    def test_argparse_usage_information_print_usage_expected_options(self):
        buf = StringIO()
        self.usage.print_usage(out=buf)
        message = buf.getvalue()
        for setting in ORDERED_SETTINGS:
            _, _, _arg = format_argparse_key(setting)
            self.assertIn(_arg, message)

    def test_argparse_usage_information_print_usage_expected_groups(self):
        buf = StringIO()
        self.usage.print_usage(out=buf)
        message = buf.getvalue()
        for group in GROUPED_SETTINGS.keys():
            if group == "_":
                group = "optional arguments"
            self.assertIn(group + ":", message)

    def test_argparse_usage_check_flag_not_specified(self):
        self.assertFalse(self.usage.check_flag(["--nope"]))

    def test_argparse_usage_check_flag_specified(self):
        with patch('omniconf.backends.argparse.ARGPARSE_SOURCE', ["--yep"]):
            self.assertTrue(self.usage.check_flag(["--yep"]))



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.


from omniconf.backends.argparse import ArgparseUsageInformation
from omniconf.setting import DEFAULT_REGISTRY as SETTING_REGISTRY
import sys


def help_requested():
    """
    Returns True if `-h` or `--help` was specified on the command line.
    """
    return flag_requested(["-h", "--help"])


def version_requested():
    """
    Returns True if `-v` or `--version` was specified on the command line.
    """
    return flag_requested(["-v", "--version"])


def flag_requested(flags):
    """
    Returns True if the specified list of flags were specified on the
    command line.
    """
    return ArgparseUsageInformation.check_flag(flags)


def show_usage(setting_registry=None, name=None, top_message=None,
               bottom_message=None, out=None, exit=0):
    """
    Prints usage information based on :class:`.Setting` objects in the given
    :class:`.SettingRegistry`. If no `setting_registry` is specified, the
    default :class:`.SettingRegistry` is used.

    If no `name` is specified, `sys.argv[0]` is used. Additionally, a header
    and footer message may be supplied using `top_message` and `bottom_message`
    message respectively.

    By default the usage information is output to `sys.stderr`. This can be
    overidden by specifying a different File-like object to `out`.

    By default, this function will call `sys.exit` and stop the program with
    exit code 0. This can be overridden by a specifying different value to
    `exit`. Set to False to not exit.
    """
    if not setting_registry:
        setting_registry = SETTING_REGISTRY

    ArgparseUsageInformation(
        setting_registry=setting_registry, name=name, top_message=top_message,
        bottom_message=bottom_message
    ).print_usage(out=out)

    if exit is not False:
        sys.exit(exit)



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

import hvac
from hvac.tests.utils import ServerManager
from mock import patch, ANY
from nose.plugins.skip import SkipTest
from omniconf.backends import available_backends
from omniconf.backends.vault import VaultBackend
from omniconf.config import ConfigRegistry
from omniconf.exceptions import InvalidBackendConfiguration
from omniconf.setting import SettingRegistry, Setting
from tempfile import NamedTemporaryFile
import nose.tools
import unittest

VAULT_CONFIG = """
backend "inmem" {
}

listener "tcp" {
address = "127.0.0.1:18200"
tls_disable = 1
}

disable_mlock = true
"""

NORMAL_POLICY = """
path "secret/*" {
    capabilities = ["read"]
}
"""

DENY_POLICY = """
path "secret/*" {
    capabilities = ["deny"]
}
"""

NORMAL_TOKEN = {
    "id": "NORMAL",
    "policies": ["normal"]
}

DENY_TOKEN = {
    "id": "DENY",
    "policies": ["deny"]
}


def test_vault_backend_in_available_backends():
    nose.tools.assert_in(VaultBackend, available_backends)


class TestVaultBackend(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        with NamedTemporaryFile() as f:
            f.write(VAULT_CONFIG.encode('utf-8'))
            f.flush()

            cls.root_client = hvac.Client(url="http://localhost:18200")
            cls.manager = ServerManager(config_path=f.name,
                                        client=cls.root_client)
            try:
                cls.manager.start()
            except OSError:  # pragma: nocover
                raise SkipTest("vault binary not present in PATH.")

            try:
                cls.manager.initialize()
                cls.manager.unseal()
                cls.root_client.token = cls.manager.root_token
                cls.root_client.set_policy("deny", DENY_POLICY)
                cls.root_client.set_policy("normal", NORMAL_POLICY)
                cls.deny_token = cls.root_client.create_token(
                    token_id="DENY",
                    policies=["deny"])['auth']['client_token']
                cls.normal_token = cls.root_client.create_token(
                    token_id="NORMAL",
                    policies=["normal"])['auth']['client_token']
            except Exception:  # pragma: nocover
                cls.manager.stop()
                raise

    @classmethod
    def tearDownClass(cls):
        cls.manager.stop()

    def setUp(self):
        self.vault = VaultBackend(url="http://localhost:18200",
                                  auth="token", credentials=self.normal_token)
        self.keys = []

    def tearDown(self):
        self.clean_keys()

    def write_key(self, key, **kwargs):
        self.root_client.write(key, **kwargs)
        self.keys.append(key)

    def clean_keys(self):
        for key in self.keys:
            self.root_client.delete(key)
        self.keys = []

    def test_vault_backend_get(self):
        self.write_key("secret/foo", bar="baz")
        setting = Setting(key="secret.foo.bar", _type=str)
        self.assertEqual(self.vault.get_value(setting), "baz")

    def test_vault_backend_get_no_node(self):
        with self.assertRaises(KeyError):
            self.vault.get_value(Setting(key="secret.foo.bar", _type=str))

    def test_vault_backend_get_no_value(self):
        self.write_key("secret/foo", bar="baz")
        with self.assertRaises(KeyError):
            self.vault.get_value(Setting(key="secret.foo.baz", _type=str))

    def test_vault_backend_get_no_access(self):
        self.write_key("secret/foo", bar="baz")
        with self.assertRaises(KeyError):
            self.vault.client.token = self.deny_token
            self.vault.get_value(Setting(key="secret.foo.bar", _type=str))

    def test_vault_backend_get_with_prefix(self):
        self.write_key("secret/foo", bar="baz")
        self.vault.prefix = "secret"
        setting = Setting(key="foo.bar", _type=str)
        self.assertEqual(self.vault.get_value(setting), "baz")

    def test_vault_backend_without_prefix_or_basepath(self):
        vault = VaultBackend(url="http://localhost:18200",
                             auth="token", credentials=self.normal_token)
        self.assertEqual(vault.prefix, "")

    def test_vault_backend_with_prefix_no_basepath(self):
        vault = VaultBackend(url="http://localhost:18200",
                             auth="token", credentials=self.normal_token,
                             prefix="foo")
        self.assertEqual(vault.prefix, "foo")

    def test_vault_backend_with_basepath_no_prefix(self):
        vault = VaultBackend(url="http://localhost:18200",
                             auth="token", credentials=self.normal_token,
                             base_path="base")
        self.assertEqual(vault.prefix, "base")

    def test_vault_backend_with_prefix_and_basepath(self):
        vault = VaultBackend(url="http://localhost:18200",
                             auth="token", credentials=self.normal_token,
                             prefix="foo", base_path="base")
        self.assertEqual(vault.prefix, "base")


def _setup_vault_autoconfig(prefix):
    settings = SettingRegistry()
    for setting in VaultBackend.autodetect_settings(prefix):
        settings.add(setting)
    return ConfigRegistry(setting_registry=settings)


def test_vault_backend_autoconfigure_invalid_auth():
    prefix = "autoconf"
    configs = _setup_vault_autoconfig(prefix)
    backend = VaultBackend.autoconfigure(configs, prefix)
    nose.tools.assert_is(backend, None)


@patch("hvac.Client", autospec=True)
def test_vault_backend_autoconfigure_token(hvac_mock):
    prefix = "autoconf"
    configs = _setup_vault_autoconfig(prefix)
    configs.set(VaultBackend._config_keys(prefix)["token"], "Bla")
    backend = VaultBackend.autoconfigure(configs, prefix)
    hvac_mock.assert_called_with(url=ANY, token="Bla")
    nose.tools.assert_is_instance(backend, VaultBackend)


@patch("hvac.Client", autospec=True)
def test_vault_backend_autoconfigure_client_cert(hvac_mock):
    prefix = "autoconf"
    configs = _setup_vault_autoconfig(prefix)
    configs.set(VaultBackend._config_keys(prefix)["tls_cert"], "CERT")
    configs.set(VaultBackend._config_keys(prefix)["tls_key"], "KEY")
    backend = VaultBackend.autoconfigure(configs, prefix)

    hvac_mock.assert_called_with(url=ANY, cert=("CERT", "KEY"))
    nose.tools.assert_is_instance(backend, VaultBackend)


@patch("hvac.Client", autospec=True)
def test_vault_backend_autoconfigure_userpass(hvac_mock):
    prefix = "autoconf"
    configs = _setup_vault_autoconfig(prefix)
    configs.set(VaultBackend._config_keys(prefix)["username"], "USER")
    configs.set(VaultBackend._config_keys(prefix)["password"], "PASS")
    backend = VaultBackend.autoconfigure(configs, prefix)

    hvac_mock.assert_called_with(url=ANY)
    hvac_mock().auth_userpass.assert_called_with("USER", "PASS")
    nose.tools.assert_is_instance(backend, VaultBackend)


@patch("hvac.Client", autospec=True)
def test_vault_backend_autoconfigure_ldap(hvac_mock):
    prefix = "autoconf"
    configs = _setup_vault_autoconfig(prefix)
    configs.set(VaultBackend._config_keys(prefix)["ldap_user"], "USER")
    configs.set(VaultBackend._config_keys(prefix)["ldap_pass"], "PASS")
    backend = VaultBackend.autoconfigure(configs, prefix)

    hvac_mock.assert_called_with(url=ANY)
    hvac_mock().auth_ldap.assert_called_with("USER", "PASS")
    nose.tools.assert_is_instance(backend, VaultBackend)


@patch("hvac.Client", autospec=True)
def test_vault_backend_autoconfigure_app_id(hvac_mock):
    prefix = "autoconf"
    configs = _setup_vault_autoconfig(prefix)
    configs.set(VaultBackend._config_keys(prefix)["app_id"], "APPID")
    configs.set(VaultBackend._config_keys(prefix)["user_id"], "USERID")
    backend = VaultBackend.autoconfigure(configs, prefix)

    hvac_mock.assert_called_with(url=ANY)
    hvac_mock().auth_app_id.assert_called_with("APPID", "USERID")
    nose.tools.assert_is_instance(backend, VaultBackend)


@patch("hvac.Client", autospec=True)
def test_vault_backend_autoconfigure_approle(hvac_mock):
    prefix = "autoconf"
    configs = _setup_vault_autoconfig(prefix)
    configs.set(VaultBackend._config_keys(prefix)["app_role"], "ROLE")
    configs.set(VaultBackend._config_keys(prefix)["app_secret"], "SECRET")
    backend = VaultBackend.autoconfigure(configs, prefix)

    hvac_mock.assert_called_with(url=ANY)
    hvac_mock().auth_approle.assert_called_with("ROLE", "SECRET")
    nose.tools.assert_is_instance(backend, VaultBackend)


@patch("hvac.Client", autospec=True)
def test_vault_backend_autoconfigure_not_authenticated(hvac_mock):
    prefix = "autoconf"
    configs = _setup_vault_autoconfig(prefix)
    configs.set(VaultBackend._config_keys(prefix)["username"], "USER")
    configs.set(VaultBackend._config_keys(prefix)["password"], "PASS")

    hvac_mock().is_authenticated.return_value = False
    with nose.tools.assert_raises(InvalidBackendConfiguration):
        VaultBackend.autoconfigure(configs, prefix)


def test_vault_backend_invalid_auth_mechanism():
    with nose.tools.assert_raises(InvalidBackendConfiguration):
        VaultBackend()



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from omniconf.backends.generic import ConfigBackend
from omniconf.keys import join_key
from omniconf.setting import Setting
import yaml


class YamlBackend(ConfigBackend):
    """
    Uses a YAML string as a backend, and allows values in it to
    be retrieved using dotted keys.
    """

    def __init__(self, conf):
        loaded_conf = {}
        for doc in yaml.full_load_all(conf):
            loaded_conf.update(doc)

        super(YamlBackend, self).__init__(loaded_conf)

    @classmethod
    def autodetect_settings(cls, autoconfigure_prefix):
        return (Setting(key=join_key(autoconfigure_prefix, "yaml", "filename"),
                        _type=str, required=False),)

    @classmethod
    def autoconfigure(cls, conf, autoconfigure_prefix):
        filename_key = join_key(autoconfigure_prefix, "yaml", "filename")
        if conf.has(filename_key):
            with open(conf.get(filename_key)) as config_file:
                return YamlBackend(conf=config_file)
        return None



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from mock import patch
from omniconf import setting, help_requested, version_requested, \
                        flag_requested, show_usage
from omniconf.setting import DEFAULT_REGISTRY as SETTING_REGISTRY
import nose.tools
import unittest


try:
    from StringIO import StringIO
except ImportError:  # pragma: nocover
    from io import StringIO


REQUESTED_TESTS = [
    (help_requested, [], None, False),
    (help_requested, ['-h'], None, True),
    (help_requested, ['--help'], None, True),
    (version_requested, [], None, False),
    (version_requested, ['-v'], None, True),
    (version_requested, ['--version'], None, True),
    (flag_requested, [], ['--nope'], False),
    (flag_requested, ['--yep'], ['--yep'], True),
]


def test_flag_methods():
    for func, cli, args, status in REQUESTED_TESTS:
        yield _test_flag_methods, func, cli, args, status


def _test_flag_methods(func, cli, args, status):
    with patch('omniconf.backends.argparse.ARGPARSE_SOURCE', cli):
        if args:
            nose.tools.assert_equal(func(args), status)
        else:
            nose.tools.assert_equal(func(), status)


class TestShowUsage(unittest.TestCase):
    def setUp(self):
        setting("foo.bar", _type=str)
        self.out = StringIO()

    def tearDown(self):
        self.out.close()
        SETTING_REGISTRY.clear()

    def test_show_usage_no_exit(self):
        show_usage(out=self.out, exit=False)

    def test_show_usage_exit(self):
        with self.assertRaises(SystemExit) as se:
            show_usage(out=self.out, exit=2)
        self.assertEqual(se.exception.code, 2)

    def test_show_usage(self):
        # Detailed tests are handled in
        # omniconf.tests.backend.test_argpparse_backend_usage
        show_usage(out=self.out, exit=False)
        self.assertIn("--foo-bar", self.out.getvalue())



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from omniconf.keys import join_key
import nose.tools

KEYS = [
    ("foo", "foo"),
    ("foo.bar", "foo", "bar"),
    ("foo.bar.baz", "foo", "bar", "baz"),
    ("foo.bar", None, "foo", None, "bar"),
    ("foo.baz", None, None, None, "foo", "baz"),
    ("foo", "foo", None, None, "", False, 0)
]


def test_join_key():
    for _tup in KEYS:
        yield _test_join_key, _tup[0], _tup[1:]


def _test_join_key(_out, parts):
    nose.tools.assert_equal(_out, join_key(*parts))



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from omniconf.backends.generic import ConfigBackend
from omniconf.keys import join_key
import os


class EnvBackend(ConfigBackend):
    """
    Uses the current process Environment, and allows values in it to
    be retrieved using dotted keys with a specific prefix. By default no
    prefix is assumed.
    """

    def __init__(self, conf=None, prefix=None):
        super(EnvBackend, self).__init__()
        self.prefix = prefix if prefix else ""

    @classmethod
    def autoconfigure(cls, conf, autoconfigure_prefix):
        return EnvBackend(prefix=conf.get(join_key(autoconfigure_prefix,
                                                   "prefix")))

    def get_value(self, setting):
        """
        Retrieves the value for the given :class:`.Setting`. Keys are converted
        as follows:

        * Dots are replaced by underscores
        * The key is uppercased.
        * A prefix is attached to the key

        This means that a key like section.value will be queried like
        ``PREFIX_SECTION_VALUE``. When no prefix is specified,
        ``SECTION_VALUE`` is queried instead.
        """
        _key = setting.key.replace(".", "_").upper()
        if self.prefix:
            _key = "{0}_{1}".format(self.prefix, _key)
        return os.environ[_key]



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from omniconf.config import config
from omniconf.setting import setting
from omniconf.loader import omniconf_load
from omniconf.usage import help_requested, version_requested, flag_requested, \
                            show_usage

__all__ = ["config", "setting", "omniconf_load", "help_requested",
           "version_requested", "flag_requested", "show_usage"]



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from omniconf.backends import available_backends, autodetection_backends
from omniconf.config import ConfigRegistry, DEFAULT_REGISTRY as CONFIG_REGISTRY
from omniconf.keys import join_key
from omniconf.setting import SettingRegistry, Setting


def autoconfigure_backends(autoconfigure_prefix=None):
    """
    Determine available backends, based on the current configuration available
    in the environment and command line. Backends can define a
    :class:`.Setting` that is required for proper autodetection.

    The result of this function is a list of backends, that are configured and
    ready to use.
    """
    if autoconfigure_prefix is None:
        autoconfigure_prefix = "omniconf"
    backend_settings = SettingRegistry()
    backend_settings.add(Setting(join_key(autoconfigure_prefix, "prefix"),
                                 _type=str))

    # Expand backend_settings with backend specific settings
    for backend in available_backends:
        _settings = backend.autodetect_settings(autoconfigure_prefix)
        if _settings:
            for _setting in _settings:
                backend_settings.add(_setting)

    # Build config for backend settings
    backend_config = ConfigRegistry(setting_registry=backend_settings)
    backend_config.load([backend() for backend in autodetection_backends])

    # Initialize backends based on detected config
    configured_backends = []
    for backend in available_backends:
        _backend = backend.autoconfigure(backend_config, autoconfigure_prefix)
        if _backend:
            configured_backends.append(_backend)

    return configured_backends


def omniconf_load(config_registry=None, backends=None,
                  autoconfigure_prefix=None):
    """
    Fill the provided :class:`.ConfigRegistry`, by default using all available
    backends (as determined by :func:`autoconfigure_backends`. If no
    :class:`ConfigRegistry` is provided, the default :class:`.ConfigRegistry`
    is used. If unset, autoconfigure_prefix will default to "omniconf".
    """
    if not config_registry:
        config_registry = CONFIG_REGISTRY
    if not backends:
        backends = autoconfigure_backends(autoconfigure_prefix)
    config_registry.load(backends)



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.



# Copyright (c) 2016 Cyso < development [at] cyso . com >
#
# This file is part of omniconf, a.k.a. python-omniconf .
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3.0 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from omniconf.backends.generic import ConfigBackend
from omniconf.exceptions import InvalidBackendConfiguration
from omniconf.keys import join_key, join_key_parts
from omniconf.setting import Setting
import hvac


class VaultBackend(ConfigBackend):
    """
    Uses Hashicorp's Vault as a backend, and allows values in it to be
    retrieved using dotted keys.

    **Key translation**

    Dotted keys are translated into an URL path, which is then optionally
    prepended by the configured backend prefix. The last part of the path is
    used as a property to retrieve. If a base_path is also configured, it
    overrides the backend prefix.

    For instance, a setting with key `setting.foo.bar` will be translated into
    path ``setting/foo``, from which the property with key ``bar`` will be
    retrieved. Because Vault nodes are grouped by backend, it usually makes
    sense to define `base_path` as `secret`, which corresponds to the Generic
    backend of Vault. In this example, the example key will be translated into
    path ``secret/setting/foo``, from which the property with key `bar` will be
    retrieved.

    **API Connection**

    The URL endpoint which |project| will default to ``http://localhost:8200``,
    and can be configured using the configuration key ``omniconf.vault.url``,
    assuming the ``autoconfigure_prefix`` is set to `omniconf`.

    **Authentication**

    Vault's API requires some form of authentication, of which the following
    are supported:

        * `Tokens <https://www.vaultproject.io/docs/auth/token.html>`_
        * `TLS certificates <https://www.vaultproject.io/docs/auth/cert.html>`_
        * `Username & Password \
           <https://www.vaultproject.io/docs/auth/userpass.html>`_
        * `LDAP <https://www.vaultproject.io/docs/auth/ldap.html>`_
        * `App ID <https://www.vaultproject.io/docs/auth/app-id.html>`_
        * `AppRole <https://www.vaultproject.io/docs/auth/approle.html>`_

    Retrieval of Vault data requires an ACL to be defined, which goes beyond
    the scope of this documentation. |project| only needs read rights on the
    keys it tries to access.

    Selection of what authentication method is used depends on which
    configuration is present during setup. For all the following examples, the
    ``autoconfigure_prefix`` is assumed to be `omniconf`:

        * Token authentication is used if ``omniconf.vault.auth.token``
          is defined.
        * TLS certificates authentication is used if both
          ``omniconf.vault.auth.tls.cert.filename`` and
          ``omniconf.vault.auth.tls.key.filename`` are defined.
        * Username and Password authentication is used if both
          ``omniconf.vault.auth.userpass.username`` and
          ``omniconf.vault.auth.userpass.password`` are defined.
        * LDAP authentication is used if both
          ``omniconf.vault.auth.ldap.username`` and
          ``omniconf.vault.auth.ldap.password`` are defined.
        * App ID authentication is used if both
          ``omniconf.vault.auth.appid.app_id`` and
          ``omniconf.vault.auth.appid.user_id`` are defined.
        * AppRole authentication is used if both
          ``omniconf.vault.auth.approle.role_id`` and
          ``omniconf.vault.auth.approle.secret_id`` are defined.

    The above order is also the order in which the configuration values are
    looked up. The first one to satisfy the conditions is used, and no further
    attepts are made if configuration fails.
    """

    def __init__(self, conf=None, prefix=None, url=None, auth=None,
                 credentials=None, base_path=None):
        if not prefix:
            prefix = ""
        if base_path:
            prefix = base_path
        self.prefix = prefix

        if auth == "token":
            self.client = hvac.Client(url=url, token=credentials)

        elif auth == "client_cert":
            self.client = hvac.Client(url=url, cert=credentials)
            self.client.auth_tls()

        elif auth in ("userpass", "ldap", "appid", "approle"):
            identifier, secret = credentials
            self.client = hvac.Client(url=url)
            if auth == "userpass":
                self.client.auth_userpass(identifier, secret)
            elif auth == "ldap":
                self.client.auth_ldap(identifier, secret)
            elif auth == "appid":
                self.client.auth_app_id(identifier, secret)
            else:
                self.client.auth_approle(identifier, secret)

        else:
            raise InvalidBackendConfiguration(
                "Invalid authentication mechanism selected for Vault backend.")

        if not self.client.is_authenticated():
            raise InvalidBackendConfiguration(
                "Vault backend is not authenticated")

    @classmethod
    def _config_keys(cls, autoconfigure_prefix):
        return {
            "url": join_key(autoconfigure_prefix, "vault", "url"),
            "base_path": join_key(autoconfigure_prefix, "vault", "base_path"),
            "token": join_key(autoconfigure_prefix, "vault", "auth", "token"),
            "tls_cert": join_key(autoconfigure_prefix, "vault", "auth", "tls",
                                 "cert", "filename"),
            "tls_key": join_key(autoconfigure_prefix, "vault", "auth", "tls",
                                "key", "filename"),
            "username": join_key(autoconfigure_prefix, "vault", "auth",
                                 "userpass", "username"),
            "password": join_key(autoconfigure_prefix, "vault", "auth",
                                 "userpass", "password"),
            "ldap_user": join_key(autoconfigure_prefix, "vault", "auth",
                                  "ldap", "username"),
            "ldap_pass": join_key(autoconfigure_prefix, "vault", "auth",
                                  "ldap", "password"),
            "app_id": join_key(autoconfigure_prefix, "vault", "auth", "appid",
                               "app_id"),
            "user_id": join_key(autoconfigure_prefix, "vault", "auth", "appid",
                                "user_id"),
            "app_role": join_key(autoconfigure_prefix, "vault", "auth",
                                 "approle", "role_id"),
            "app_secret": join_key(autoconfigure_prefix, "vault", "auth",
                                   "approle", "secret_id")
        }

    @classmethod
    def autodetect_settings(cls, autoconfigure_prefix):
        settings = []
        for name, key in cls._config_keys(autoconfigure_prefix).items():
            if name == "url":
                settings.append(Setting(key=key, _type=str, required=False,
                                        default="http://localhost:8200"))
            else:
                settings.append(Setting(key=key, _type=str, required=False))
        return settings

    @classmethod
    def autoconfigure(cls, conf, autoconfigure_prefix):
        keys = cls._config_keys(autoconfigure_prefix)
        url = conf.get(keys['url'])

        if conf.has(keys['token']):
            return VaultBackend(url=url, auth="token",
                                credentials=conf.get(keys['token']))

        elif conf.has(keys['tls_cert']) and conf.has(keys['tls_key']):
            return VaultBackend(url=url, auth="client_cert",
                                credentials=(conf.get(keys['tls_cert']),
                                             conf.get(keys['tls_key'])))

        elif conf.has(keys['username']) and conf.has(keys['password']):
            return VaultBackend(url=url, auth="userpass",
                                credentials=(conf.get(keys['username']),
                                             conf.get(keys['password'])))

        elif conf.has(keys['ldap_user']) and conf.has(keys['ldap_pass']):
            return VaultBackend(url=url, auth="ldap",
                                credentials=(conf.get(keys['ldap_user']),
                                             conf.get(keys['ldap_pass'])))

        elif conf.has(keys['app_id']) and conf.has(keys['user_id']):
            return VaultBackend(url=url, auth="appid",
                                credentials=(conf.get(keys['app_id']),
                                             conf.get(keys['user_id'])))

        elif conf.has(keys['app_role']) and conf.has(keys['app_secret']):
            return VaultBackend(url=url, auth="approle",
                                credentials=(conf.get(keys['app_role']),
                                             conf.get(keys['app_secret'])))

        return None

    def get_value(self, setting):
        parts = setting.key.split(".")
        if self.prefix:
            parts.insert(0, self.prefix)

        data_key = parts.pop()
        path = join_key_parts("/", parts)

        try:
            node = self.client.read(path)
        except hvac.exceptions.Forbidden:
            raise KeyError("No access to Vault data node at {0}".format(path))
        if not node:
            raise KeyError("No Vault data node at {0}".format(path))
        return node['data'][data_key]
